'use strict'

const   https                   = require( 'https' )
const   http                    = require( 'http' )
const   path                    = require( 'path' )
const   fs                      = require( 'fs' )
const   secretary               = require( '@clerksystems/secretary' )
const   _                       = require( 'lodash' )
const   uuid                    = require( 'uuid/v4' )
const   WebSocket               = require( 'ws' )
const   lib                     = require( path.join( __dirname, 'lib.js' ) )
const   ConnectionWrapper       = require( path.join( __dirname, 'ConnectionWrapper.js' ) )

const   internalPlugin          = require( path.join( __dirname, '..', 'plugins', 'yoctopus.js' ) )
const   api                     = require( path.join( __dirname, '..', 'plugins', 'api.js' ) )

const   bundledPlugins          = {
                                    multicast   : require( path.join( __dirname, '..', 'plugins', 'multicast.js' ) )
                                // ,   'mini-auth' : require( path.join( __dirname, '..', 'plugins', 'mini-auth.js' ) )
                                }
const	{value, toObject}		= require('@clerksystems/concise-schema');

const   pause                   = s => new Promise( res => setTimeout( res, s * 1000 ))





const ServerWrapper = function(
    {
        staticEndpointArguments = {}        // an object with arguments that gets supplied to endpoint request handlers on requests
    ,   port                    = 4000
    ,   keyPath                 = null
	,   certPath                = null
    ,   caPath                  = null
	,   logger                  = null
    ,   test                    = false
    ,   logInstructions         = {
                                    'info' : 'console'
                                }
	,	allowedOrigins          = '*'
    ,   onError                 = () => {}
    ,   pingTime
    ,   pingFailAmount
    ,   createApi               = false
    } = {})
{
    const   server              = {}
    const   id                  = uuid().substr( 0, 4 )
    let     started             = false

    const   mode                = test === false ? 'production' : 'test'

    const   logListeners        = {}

    let     logRegister         = []

    const   notifyLogFn         = function( logObj )
                                {
                                    // console.log( 'logObj', logObj)
                                    logObj.id           = uuid()
                                    logObj.logTime      = new Date().getTime()

                                    logRegister = _.take( [ logObj ].concat( logRegister ), 250 )

                                    Object
                                        .values( logListeners )
                                        .forEach(
                                            listener =>
                                                listener( logObj )
                                        )
                                }

    // const   logInfoInstruction  = logInstructions.info
    //                             ?   logInstructions.info
    //                             :   () => ''

    const   augmLogInfoInstr    = []

    if ( logInstructions.info )
    {
        augmLogInfoInstr.push( logInstructions.info )
    }

    augmLogInfoInstr.push( notifyLogFn )

    const   augmLogInstructions = Object.assign(
                                    {}
                                ,   logInstructions
                                ,   {
                                        info    : augmLogInfoInstr
                                    }
                                )

    const   makeLogger          = ( source, instructions ) =>
                                    logger
                                    ?   logger
                                    :   secretary
                                            .create({
                                                debug : false
                                            ,   source
                                            ,   instructions
                                            })

    const   log                 = makeLogger( 'yoctopus', augmLogInstructions )

    if ( ! log.instructions )
    {
        throw new Error( 'Logger must be secretary v1.0.1 or higher' )
    }

    // log.reInstruct(
        // Object.assign(
        //     {}
        // ,   log.instructions
        // ,   {
        //         info    : function( logObj )
        //                 {
        //                     logObj.id           = uuid()
        //                     logObj.logTime      = new Date().getTime()

        //                     logRegister = _.take( [ logObj ].concat( logRegister ), 250 )

        //                     Object
        //                         .values( logListeners )
        //                         .forEach(
        //                             listener =>
        //                                 listener( logObj )
        //                         )
        //                 }
        //     }
        // )
    // )

    const   subscribeToLogs     = function( listener )
                                {
                                    const   listenerId  = uuid()
                                    const   rmListener  = function()
                                                        {
                                                            delete logListeners[ listenerId ]
                                                        }

                                    logListeners[ listenerId ] = listener

                                    return rmListener
                                }

    // staticEndpointArguments.log = log



    const   indexedEndpoints    = {}

    const   pluginComponents    = {
                                        endpoints       : {}
                                    ,   handlers        : {
                                                            onConnectionOpened          : []
                                                        ,   onConnectionChanged         : []
                                                        ,   onConnectionClosed          : []
                                                        ,   onRequestReceived           : []
                                                        ,   onResponsePrepared          : []
                                                        ,   onResponseSent              : []
                                                        }
                                    ,   names           : []
                                    ,   settings        : {}
                                    ,   listeners       : {}
                                    }



    const   connectionRegister  = {}


    const   plainServer         = keyPath
                                ?   https.createServer({
                                        cert: fs.readFileSync( certPath ),
                                        key: fs.readFileSync( keyPath ),
										ca: fs.readFileSync( caPath ),
                                    })

                                :   http.createServer()

    lib.enableDestroy( plainServer )

    const   wsServer            = new WebSocket.Server({ server: plainServer })
    wsServer.on('error', onError)


    Object.defineProperty(
        server
    ,   'logs'
    ,   { get() { return logRegister } }
    )


    server.port                 = port
    server.isSecure             = !! keyPath
    server.logger               = log
    server.subscribeToLogs      = subscribeToLogs

    server.plainServer          = plainServer

    server.connectionRegister   = connectionRegister

    server.pluginComponents     = pluginComponents
    server.indexedEndpoints     = indexedEndpoints

    server.bundledPlugins       = bundledPlugins

    server.close                = async function()
                                {
                                    let     closeRes        = null
                                    const   closePromise    = new Promise( res => closeRes = res )

                                    plainServer.destroy( function()
                                    {
                                        log.notice( `Server ${id} closed` )
                                        closeRes()
                                    })

                                    return closePromise
                                }


    // overwrites itself and server.plugin on call
    // might add a check that matches listeners to settings

    server.start               = async function()
                                {
                                    started                 = true

                                    let     listenRes       = null
                                    const   listenPromise   = new Promise( res => listenRes = res )

                                    for ( let [ endpointName, endpointData ] of Object.entries( pluginComponents.endpoints ) )
                                    {

                                        // save the scopeBuilder etc to be used in every method/endpoint combo
                                        const staticBits = _.omit( endpointData, 'handlers' )

                                        for ( let [ method, handlerObj ] of Object.entries( endpointData.handlers ) )
                                        {
                                            indexedEndpoints[ method + ':' + endpointName ] = {

                                                // this is the handler object as defined in the plugin
                                                handler         : handlerObj.handler
                                            ,   requirements    : handlerObj.requirements   || null
                                            ,   requestFormat   : handlerObj.requestFormat  || null

                                            ,   ... staticBits

                                            }
                                        }
                                    }

                                    Object.freeze( indexedEndpoints )
                                    Object.freeze( pluginComponents )
                                    Object.freeze( staticEndpointArguments )

                                    server.id      = id
                                    server.server  = wsServer

                                    wsServer.on(
                                        'listening'
                                    ,   function()
                                        {
                                            log.info( `Server ${id} listening` )
                                            listenRes()
                                        }
                                    )

                                    plainServer.listen( port )

                                    server.start   = () => { throw new Error( 'Server already started' ) }
                                    server.plugin  = () => { throw new Error( 'Server already started' ) }

                                    return listenPromise
                                }

    // server.multicast            = function( filter, message )
    //                             {
    //                                 server.connectionRegister()
    //                             }

    server.plugin               = function(
                                    {
                                        name
                                    ,   listeners               = {}
                                    ,   endpoints               = {}
                                    ,   settings                = {}
                                    ,   passToListeners         = null
                                    } = {}
                                ,   {
                                        endpointAliases = {}
                                    } = {}
                                )
                                {
                                    if ( typeof( name ) !== 'string' || ! name )
                                    {
                                        throw new Error( 'Plugin must have a name.' )
                                    }

                                    log.notice( `Installing plugin "${name}"` )

                                    if ( Object.keys( pluginComponents.settings ).indexOf( name ) !== -1 )
                                    {
                                        throw new Error( `Plugin must have a unique name. Server already has a plugin named ${name}.` )
                                    }

                                    const   id              = uuid().substr( 0, 8 )

                                    const   pluginLogger    = makeLogger( name, augmLogInstructions )

                                    pluginComponents.settings[ name ] = {}

                                    if ( passToListeners )
                                    {
                                        staticEndpointArguments[ name ] = passToListeners
                                    }

                                    // load endpoints

                                    const   epPrefix        = `${name}:`

                                    for ( let endpointName of Object.keys( endpoints ) )
                                    {
										for (const methodName in endpoints[endpointName]) {
											const handlerObj = endpoints[endpointName][methodName];
											if(typeof handlerObj.requestFormat === 'function' || handlerObj.requestFormat === undefined) {
												toObject(value.any);
											} else {
												handlerObj.requestFormat = lib.convertStringObject(handlerObj.requestFormat)
												toObject(handlerObj.requestFormat);
											}
										}
                                        const   prefixedEpName  = `${epPrefix}${endpointName}`
                                        const   endpointObj     = {
                                                                    handlers            : endpoints[ endpointName ]
                                                                ,   pluginId            : id
                                                                ,   handleScopeBuilder  : buildHandleScope( name )
                                                                ,   logger              : pluginLogger
                                                                ,   name
                                                                }

                                        pluginComponents.endpoints[ prefixedEpName ] = endpointObj

                                        if ( endpointAliases[ endpointName ] !== undefined )
                                        {
                                            // user intends to map this plugin endpoint to the
                                            // global namespace

                                            const   globalEpName    = endpointAliases[ endpointName ]

                                            if ( pluginComponents.endpoints[ globalEpName ] !== undefined )
                                            {
                                                throw new Error( `Cannot map plugin ${name} endpoint ${prefixedEpName} to existing endpoint ${globalEpName}` )
                                            }

                                            pluginComponents.endpoints[ globalEpName ] = endpointObj
                                        }
                                    }

                                    for ( let handlerName of Object.keys( pluginComponents.handlers ) )
                                    {
                                        if ( listeners.server && listeners.server[ handlerName ] )
                                        {
                                            log.debug( `Using handler ${handlerName}` )

                                            pluginComponents
                                                .handlers[ handlerName ]
                                                .push({
                                                    name
                                                ,   id
                                                ,   handleScopeBuilder  : buildHandleScope( name )
                                                ,   fn                  : listeners.server[ handlerName ]
                                                ,   logger              : pluginLogger
                                                })
                                        }
                                    }

                                    for ( let [ setting, type ] of Object.entries( settings ) )
                                    {
                                        pluginComponents.settings[ name ][ setting ] = type
                                    }

                                    if ( listeners.plugins )
                                    {
                                        for ( let [ targetPlugin, targetSettingsOrWCHandler ] of Object.entries( listeners.plugins ) )
                                        {
                                            if ( targetPlugin === '*' )
                                            {
                                                addListener( targetSettingsOrWCHandler, '*' )
                                            }
                                            else
                                            {
                                                for ( let [ setting, listener ] of Object.entries( targetSettingsOrWCHandler ) )
                                                {
                                                    addListener( listener, [ targetPlugin, setting ] )
                                                }
                                            }
                                        }
                                    }

                                    log.debug( `Installed plugin "${name}"` )

                                    return server
                                }

    const   addListener         = function( listener, where )
                                {
                                    const currentListeners = _.get( pluginComponents.listeners, where, [] )
                                    _.set( pluginComponents.listeners, where, currentListeners.concat( listener ) )
                                }


    const   buildHandleScope    = pluginName =>
                                    connectionHandles =>
                                        ({
                                            ...connectionHandles.getter
                                        ,   ...connectionHandles.setters[ pluginName ]
                                        })


	Object.defineProperty(
		server
	,   'allowedOrigins'
	,	{
			get()
			{
				return allowedOrigins
			}
		,	set( newOrigins )
			{
                if ( ! Array.isArray( newOrigins ) && newOrigins !== '*')
                {
                    throw new Error( 'Allowed origins should be an array or string "*"' )
                }
				allowedOrigins = newOrigins
			}
		}
	)

    wsServer.on(
        'connection'
    ,   async function( rawConnection, request )
        {
            const   remoteAddress       = request.headers['x-forwarded-for'] || request.connection.remoteAddress

            log.info( `Connection opened from ${remoteAddress}` )

			// check allowed origins
			if ( allowedOrigins !== '*' && ! allowedOrigins.includes( request.headers.origin ) )
            {
                log.error( `Origin [${request.headers.origin}] not allowed` )
                rawConnection.close()
                return false
            }

			// start the connection
            rawConnection.remoteAddress = remoteAddress

            const   wrappedConnection   = new ConnectionWrapper({
                                            connection      : rawConnection
                                        ,   endpoints       : indexedEndpoints
                                        ,   pluginSettings  : _.cloneDeep( pluginComponents.settings )
                                        ,   serverId        : id
                                        ,   handlers        : pluginComponents.handlers
                                        ,   staticEndpointArguments
                                        ,   listeners       : pluginComponents.listeners
                                        ,   log
                                        ,   onClose         : function( connectionId )
                                                            {
																connectionRegister[ connectionId ] = undefined;
                                                                delete connectionRegister[ connectionId ]
                                                            }
                                        ,   onError
                                        ,   pingTime
                                        ,   pingFailAmount
                                        })

            connectionRegister[ wrappedConnection.id ] = wrappedConnection

        }
    )

    server.plugin( internalPlugin( server ) ) // "yoctopus"
    if(createApi) {
        server.plugin( api(server), { endpointAliases: { 'api': 'api' } } )
    }

    return server
}

module.exports              = ServerWrapper
