const   path        = require( 'path' )
const   expect      = require( 'chai' ).expect
const   WebSocket   = require( 'ws' )
const   Yoctopus    = require( path.join( __dirname, '..', 'index.js' ) )
const	{value}		= require('@clerksystems/concise-schema');

const   pause       = s => new Promise( res => setTimeout( res, s * 1000 ))

const   npFact      = ( startPort = 5000 + Math.round( Math.random() * 1000 ) ) =>
                        () =>
                        {
                            startPort += 10
                            return startPort
                        }

const   nextPort    = npFact()

describe( 'Server', function()
{
    this.timeout( 3 * 1000 )

    let server              = null
    let port                = null
    let wsA                 = null
    let wsB                 = null
    let resolveTestPromise  = null
    let rejectTestPromise   = null
    let testPromise         = null

    beforeEach( async function()
    {
        resolveTestPromise  = null
        testPromise     = new Promise( (res, rej) => {resolveTestPromise = res; rejectTestPromise = rej;} )

        port    = nextPort()
        server  = new Yoctopus({ port, logInstructions: {} })
    })

    afterEach( async function()
    {
        try
        {
            await server.close()
        }
        catch( err ) { }

        try
        {
            await wsA.close()
        }
        catch( err ) {}

        try
        {
            await wsB.close()
        }
        catch( err ) {}
    })

    it( 'loads', () =>
    {
        server = new Yoctopus({logInstructions: {}})

        console.log( Yoctopus.prototype )
        // console.log( Yoctopus.promisify( console.log ) )

        // expect( Yoctopus.promisify ).to.be.a( 'function' )
        // expect( server.promisify ).to.be.a( 'function' )
    })


    it( 'accepts connections', async () =>
    {
        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   function open()
            {
                resolveTestPromise()
            }
        )

        await testPromise
    })


    it( 'refuses messages without meta entry' )

    it( 'requires unique plugin names' )


    it( 'installs plugins and calls their handlers and endpoints', async () =>
    {
        const   testLog = []

        await server.plugin({
            name        : 'test-plugin'
        ,   listeners   : {
                            server: {
                                onConnectionOpened      : function()
                                                        {
                                                            testLog.push( 'opened' )
                                                        }
                            ,   onRequestReceived       : function()
                                                        {
                                                            testLog.push( 'request received' )
                                                        }
                            ,   onResponsePrepared      : function()
                                                        {
                                                            testLog.push( 'response prepared' )
                                                        }
                            ,   onResponseSent          : function()
                                                        {
                                                            testLog.push( 'response sent' )
                                                        }
                            ,   onConnectionClosed      : function()
                                                        {
                                                            testLog.push( 'closed' )
                                                            expect( testLog ).to.deep.equal( [ 'opened', 'request received', 'ep read', 'response prepared', 'response sent', 'closed' ] )
                                                            resolveTestPromise()
                                                        }
                            }
                        }
        ,   endpoints   : {
                            test : {
                                read: {
                                    handler : async function()
                                            {
                                                testLog.push( 'ep read' )
                                            }
                                }
                            }
                        }
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   async function()
            {
                wsA.send(
                    JSON.stringify(
                        {
                            meta: { id: 'test1', method: 'read', endpoint: 'test-plugin:test' }
                        }
                    )
                )
                await pause( 0.5 )
                wsA.close()
            }
        )

        await testPromise
    })


    it( 'calls multiple plugins in order, passes on the data, static args and con settings in the right way', async () =>
    {
        const   testLog         = []
        let     result          = null
        let     checkSettings   = null

        for ( let a = 0;  a < 10;  a += 1 )
        {
            await server.plugin({
                name        : 'test-plugin ' + a
            ,   settings    : {
                                henk: 'string'
                            }
            ,   listeners   : {
                                server: {
                                    onConnectionOpened      : function({ data, connectionSettings })
                                                            {
                                                                testLog.push( 'opened ' + a )
                                                                connectionSettings.set( 'henk', 'kees' )
                                                                result = { 'test': ( data.test || '' ) + a }
                                                                return result
                                                            }

                                ,   onConnectionClosed      : function({ connectionSettings })
                                                            {
                                                                checkSettings = connectionSettings.get( 'test-plugin ' + a, 'henk' )
                                                            }
                                }
                            }
            })
        }

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   async function open()
            {
                await pause( 0.1 )

                expect( testLog )
                    .to.deep
                    .equal([
                        'opened 0',
                        'opened 1',
                        'opened 2',
                        'opened 3',
                        'opened 4',
                        'opened 5',
                        'opened 6',
                        'opened 7',
                        'opened 8',
                        'opened 9'
                    ])

                expect( result ).to.deep.equal({ test: '0123456789' })

                wsA.close()

                await pause( 0.5 )

                await server.close()

                await pause( 0.5 )

                expect( checkSettings ).to.equal( 'kees' )

                resolveTestPromise()
            }
        )

        await testPromise
    })


    it( 'responds to requests', async () =>
    {
        const rVal = Math.random()

        await server.plugin({
            name        : 'test-plugin2'
        ,   endpoints   : {
                            randomval : {
                                read: {
                                    handler : async function()
                                            {
                                                return rVal
                                            }
                                }
                            }
                        }
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   function open()
            {
                wsA.send( JSON.stringify({
                    meta: {
                        id          : 'testMsg1'
                    ,   method      : 'read'
                    ,   endpoint    : 'test-plugin2:randomval'
                    }
                }))
            }
        )

        wsA.on(
            'message'
        ,   function receive( msg )
            {
                const msgObj = JSON.parse( msg )
                expect( msgObj.data ).to.eq( rVal )
                resolveTestPromise()
            }
        )

        await testPromise
    })


    it( 'returns errors for internal server errors', async () =>
    {
       await server.plugin({
            name        : 'test-plugin4'
        ,   endpoints   : {
                            err : {
                                read: {
                                    handler : async function()
                                            {
                                                throw new Error( 'This is a fake error' )
                                            }
                                }
                            }
                        }
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   () =>  wsA.send( JSON.stringify({
                meta: {
                    id          : 'testMsgId-' + Math.random()
                ,   method      : 'read'
                ,   endpoint    : 'test-plugin4:err'
                }
            }))
        )

        wsA.on(
            'message'
        ,   function receive( msg )
            {
                const msgObj = JSON.parse( msg )
                expect( msgObj.data ).to.be.undefined
                expect( msgObj.errors[ 0 ].detail ).to.be.a.string
                expect( msgObj.errors[ 0 ].detail ).to.contain( 'fake' )
                resolveTestPromise()
            }
        )

        await testPromise
    })


    it( 'returns errors for non existant endpoints', async () =>
    {
        await server.plugin({
            name        : 'test-plugin3'
        ,   endpoints   : {
                            ok : {
                                read: {
                                    handler : async function()
                                            {
                                                return 'ok'
                                            }
                                }
                            }
                        }
        })

        await server.start()

        wsA      = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   () => wsA.send(
                JSON.stringify({
                    meta: {
                        id          : 'testMsgId-' + Math.random()
                    ,   endpoint    : 'test-plugin2:okdif'
                    ,   method      : 'read'
                    }
                })
            )
        )

        wsA.on(
            'message'
        ,   function receive( msg )
            {
                const msgObj = JSON.parse( msg )
                expect( msgObj.data ).to.be.undefined
                expect( msgObj.errors[ 0 ].detail ).to.be.a.string
                expect( msgObj.errors[ 0 ].detail ).to.contain( 'Unknown' )
                resolveTestPromise()
            }
        )

        await testPromise
    })



    it( 'allows the use of aliases', async () =>
    {
       await server.plugin(
           {
                name        : 'test-plugin-alias'
            ,   endpoints   : {
                                ok : {
                                    read: {
                                        handler : async function()
                                                {
                                                    return 'ok'
                                                }
                                    }
                                }
                            }
            ,
            }
        ,   {
                endpointAliases: {
                    'ok': 'whaaat'
                }
            }
        )

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   () => wsA.send(
                JSON.stringify({
                    meta: {
                        id          : 'testMsgId-' + Math.random()
                    ,   method      : 'read'
                    ,   endpoint    : 'whaaat'
                    }
                })
            )
        )

        wsA.on(
            'message'
        ,   function receive( msg )
            {
                const msgObj = JSON.parse( msg )
                expect( msgObj.data ).to.equal( 'ok' )
                resolveTestPromise()
            }
        )

        await testPromise
    })

    it( 'Does not allow for a duplicate alias', async () =>
    {
       await server.plugin(
           {
                name        : 'test-plugin-alias'
            ,   endpoints   : {
                                ok : {
                                    read: {
                                        handler : async function()
                                                {
                                                    return 'ok 1'
                                                }
                                    }
                                }
                            }
            ,
            }
        ,   {
                endpointAliases: {
                    'ok': 'whaaat'
                }
            }
        )

		try {

           await server.plugin(
               {
                    name        : 'test-plugin-alias-2'
                ,   endpoints   : {
                                    ok : {
                                        read: {
                                            handler : async function()
                                                    {
                                                        return 'ok 2'
                                                    }
                                        }
                                    }
                                }
                ,
                }
            ,   {
                    endpointAliases: {
                        'ok': 'whaaat'
                    }
                }
            )

		} catch( err ) {
			expect( err.message ).to.contain( 'Cannot map plugin test-plugin-alias-2 endpoint test-plugin-alias-2:ok to existing endpoint whaaat' );
		}
    })



    it( 'allows plugins to share information', async () =>
    {
        await server.plugin({
            name        : 'test-share-A'
        ,   settings    : {
                            'henk' : 'any'
                        }
        ,   endpoints   : {
                            setval : {
                                read: {
                                    handler : async function({ connectionSettings })
                                            {
                                                expect( connectionSettings.get ).to.be.a( 'function' )
                                                expect( connectionSettings.set ).to.be.a( 'function' )

                                                connectionSettings.set( 'henk', 'piet' )
                                            }
                                }
                            }
                        }
        })

        await server.plugin({
            name        : 'test-share-B'
        ,   endpoints   : {
                            getval : {
                                read: {
                                    handler : async function({ connectionSettings })
                                            {
                                                expect( connectionSettings.get ).to.be.a( 'function' )
                                                expect( connectionSettings.set ).to.be.a( 'function' )

                                                return connectionSettings.get( 'test-share-A', 'henk' )
                                            }
                                }
                            }
                        }
        })

        await server.start()

        wsA             = new WebSocket( `ws://localhost:${port}` )

        const   send    = ( endpoint, data, method = 'read' ) =>
                            wsA.send( JSON.stringify({
                                meta: {
                                    id          : 'testMsgId-' + Math.random()
                                ,   method
                                ,   endpoint
                                }
                            }))

        wsA.on(
            'open'
        ,   () => send( 'test-share-A:setval' )
        )

        let count = 0

        wsA.on(
            'message'
        ,   function receive( msg )
            {
                count ++
                switch( count )
                {
                    case 1:
                        send( 'test-share-B:getval' )
                        break

                    case 2:
                        expect( JSON.parse( msg ).data ).to.equal( 'piet')
                        resolveTestPromise()
                        break
                }
            }
        )

        await testPromise
    })


    it( 'can enforce connection requirements', async () =>
    {
        const   rVal    = Math.random()

        await server.plugin({
            name        : 'test-req'
        ,   settings    : {
                            readAllowed: 'bool'
                        }
        ,   endpoints   : {
                            readval : {
                                read: {
                                    requirements : {
                                        'test-req.readAllowed' : true
                                    }
                                ,   handler : async function({ connectionSettings })
                                            {
                                                return rVal
                                            }
                                }
                            ,   update: {
                                    handler : function({ connectionSettings })
                                            {
                                                connectionSettings.set( 'readAllowed', true )
                                                return true
                                            }
                                }
                            }
                        }
        })

        await server.start()

        wsA             = new WebSocket( `ws://localhost:${port}` )

        const   send    = ( endpoint, data, method = 'read' ) =>
                            wsA.send( JSON.stringify({
                                meta: {
                                    id          : 'testMsgId-' + Math.random()
                                ,   method
                                ,   endpoint
                                }
                            }))

                            wsA.on(
            'open'
        ,   () => send( 'test-req:readval' )
        )

        let count = 0

        wsA.on(
            'message'
        ,   function receive( msg )
            {
                count ++

                const parsedMsg = JSON.parse( msg )

                switch( count )
                {
                    case 1:
                        expect( parsedMsg.errors ).to.be.an( 'array' )
                        expect( parsedMsg.errors[ 0 ].detail ).to.include( 'requirements' )
                        send( 'test-req:readval', null, 'update' )
                        break

                    case 2:
                        expect( parsedMsg.data ).to.equal( true )
                        send( 'test-req:readval' )
                        break

                    case 3:
                        expect( parsedMsg.data ).to.equal( rVal )
                        resolveTestPromise()
                        break
                }
            }
        )

        await testPromise
    })




    it( 'can enforce a request format', async () =>
    {
        const   rVal    = Math.random()

        await server.plugin({
            name        : 'test-req-form'
        ,   endpoints   : {
                            readval : {
                                read: {
                                    requestFormat : 'string'
                                ,   handler : async function({ connectionSettings })
                                            {
                                                return rVal
                                            }
                                }
                            ,   create: {
                                    requestFormat : { 'some-entry' : 'email' }
                                ,   handler : async function({ connectionSettings })
                                            {
                                                return rVal
                                            }
                                }
                            ,   pw: {
                                    requestFormat : a => a === 'super_secret_password_code'
                                ,   handler : async function({ connectionSettings })
                                            {
                                                return rVal
                                            }
                                }
                            ,   complex: {
                                    requestFormat : {
                                        a: {
                                            b: {
                                                c: 'string'
                                            }
                                        },
                                        d: {
                                            e: 'string'
                                        }
                                    }
                                ,   handler: async function({ connectionSettings })
                                            {
                                                return rVal;
                                            }
                                }
							,	validator: {
									requestFormat: {
										a: {
											b: value.any
										},
										c: value.boolean.required
									}
								,	handler: async function({ connectionSettings }) {
										return rVal;
									}
								}
							,	arrayValidator: {
									requestFormat: value.array.like(value.object)
								,	handler: async function({ connectionSettings }) {
										return rVal;
									}
								}
							,	objectValidator: {
									requestFormat: {
										value: value.string
									}
								,	handler: async function({ connectionSettings }) {
										return rVal;
									}
								}
                            }
                        }
        })

        await server.start()

        wsA             = new WebSocket( `ws://localhost:${port}` )

        const   send    = ( endpoint, data, method = 'read' ) =>
                            wsA.send( JSON.stringify({
                                meta: {
                                    id          : 'testMsgId-' + Math.random()
                                ,   method
                                ,   endpoint
                                }
                            ,   data
                            }))

        wsA.on(
            'open'
        ,   () => send( 'test-req-form:readval' )
        )

        let count = 0

        wsA.on(
            'message'
        ,   function receive( msg )
            {
                count ++

                const parsedMsg = JSON.parse( msg )

                // console.log( parsedMsg.errors )

                switch( count )
                {
                    case 1:
                        expect( parsedMsg.errors ).to.be.an( 'array' )
                        expect( parsedMsg.errors[ 0 ].detail ).to.include( 'expected' )
                        send( 'test-req-form:readval', 'hoi', 'read' )
                        break

                    case 2:
                        expect( parsedMsg.data ).to.equal( rVal )
                        send( 'test-req-form:readval', 'hoi', 'create' )
                        break

                    case 3:
                        expect( parsedMsg.errors ).to.be.an( 'array' )
                        expect( parsedMsg.errors[ 0 ].detail ).to.include( 'expected' )
                        send( 'test-req-form:readval', { 'some-entry': 'not-an-email' }, 'create' )
                        break

                    case 4:
                        expect( parsedMsg.errors ).to.be.an( 'array' )
                        expect( parsedMsg.errors[ 0 ].detail ).to.include( 'expected' )
                        send( 'test-req-form:readval', { 'some-entry': 'test@email.com' }, 'create' )
                        break

                    case 5:
                        expect( parsedMsg.data ).to.equal( rVal )
                        send( 'test-req-form:readval', 'wrong pass', 'pw' )
                        break

                    case 6:
                        expect( parsedMsg.errors ).to.be.an( 'array' )
                        expect( parsedMsg.errors[ 0 ].detail ).to.include( 'rejected' )
                        send( 'test-req-form:readval', 'super_secret_password_code', 'pw' )
                        break

                    case 7:
                        expect( parsedMsg.data ).to.equal( rVal )
                        send( 'test-req-form:readval', {a:{b:{c:5}},d:{e:{f:''}}}, 'complex' )
                        break
                    case 8:
                        expect( parsedMsg.errors ).to.be.an( 'array' )
                        expect( parsedMsg.errors.length ).to.equal(2);
                        expect( parsedMsg.errors[ 0 ].detail ).to.include( 'expected' )
                        expect( parsedMsg.errors[ 1 ].detail ).to.include( 'expected' )
                        send( 'test-req-form:readval', {a:{b:{c:5}},d:{e:''}}, 'complex' )
                        break
                    case 9:
                        expect( parsedMsg.errors ).to.be.an( 'array' )
                        expect( parsedMsg.errors.length ).to.equal(1);
                        expect( parsedMsg.errors[ 0 ].detail ).to.include( 'expected' )
                        send( 'test-req-form:readval', {a:{b:{c:''}},d:{e:''}}, 'complex' )
                        break
                    case 10:
                        expect( parsedMsg.data ).to.equal( rVal )
						send( 'test-req-form:readval', {}, 'validator' )
						break;
					case 11:
                        expect( parsedMsg.errors ).to.be.an( 'array' );
						expect( parsedMsg.errors.length ).to.equal( 2 );
                        expect( parsedMsg.errors[ 0 ].detail ).to.include( 'expected a "object", got "undefined", at path ".a"' )
                        expect( parsedMsg.errors[ 1 ].detail ).to.include( 'expected a "boolean", got "undefined", at path ".c"' )
						send( 'test-req-form:readval', {a:{b:true}, c: true}, 'validator' )
						break;
					case 12:
						expect( parsedMsg.data ).to.equal( rVal )
						send( 'test-req-form:readval', [{}], 'arrayValidator' )
                        break
					case 13:
						// console.log( parsedMsg );
						// expect success
						send( 'test-req-form:readval', [''], 'arrayValidator' )
                        break
					case 14:
						expect( parsedMsg.errors.length ).to.equal( 1 );
						expect( parsedMsg.errors[ 0 ].detail ).to.include( 'expected a "object", got "", at path "[0]"' );
						send( 'test-req-form:readval', { value: 'value' }, 'objectValidator' )
                        break
					case 15:
						send( 'test-req-form:readval', { value: 123 }, 'objectValidator' )
                        break
					case 16:
						expect( parsedMsg.errors.length ).to.equal( 1 );
						expect( parsedMsg.errors[ 0 ].detail ).to.include( 'expected a "string", got "123", at path ".value"' );
						// expect error
						resolveTestPromise()
						break
                }
            }
        )

        await testPromise
    })

	it( 'creates a readale api', async () => {
        const   rVal    = Math.random()

		server  = new Yoctopus({ port, createApi: true, logInstructions: {} })

        await server.plugin({
            name        : 'test-req-form'
        ,   endpoints   : {
                            readval : {
                                read: {
                                    requestFormat : 'string'
                                ,   handler : async function({ connectionSettings })
                                            {
                                                return rVal
                                            }
                                }
                            ,   create: {
                                    requestFormat : { 'some-entry' : 'email' }
                                ,   handler : async function({ connectionSettings })
                                            {
                                                return rVal
                                            }
                                }
                            ,   pw: {
                                    requestFormat : a => a === 'super_secret_password_code'
                                ,   handler : async function({ connectionSettings })
                                            {
                                                return rVal
                                            }
                                }
                            ,   complex: {
                                    requestFormat : {
                                        a: {
                                            b: {
                                                c: 'string'
                                            }
                                        },
                                        d: {
                                            e: 'string'
                                        }
                                    }
                                ,   handler: async function({ connectionSettings })
                                            {
                                                return rVal;
                                            }
                                }
							,	validator: {
									requestFormat: {
										a: {
											b: value.any
										},
										c: value.boolean.required
									}
								,	handler: async function({ connectionSettings }) {
										return rVal;
									}
								}
                            }
                        }
        })

        await server.start()

        wsA             = new WebSocket( `ws://localhost:${port}` )

        const   send    = ( endpoint, data, method = 'read' ) =>
                            wsA.send( JSON.stringify({
                                meta: {
                                    id          : 'testMsgId-' + Math.random()
                                ,   method
                                ,   endpoint
                                }
                            ,   data
                            }))

                            wsA.on(
            'open'
        ,   () => send( 'api' )
        )

        wsA.on(
            'message'
        ,   function receive( message )
            {
                const msg = JSON.parse( message )

				try {
					expect(msg).to.be.an('Object');
					expect(msg.meta).to.be.an('Object');
					expect(msg.meta.endpoint).to.equal('api');
					expect(msg.meta.method).to.equal('read');
					expect(msg.data).to.be.an('Object');
					expect(msg.data['test-req-form:readval']).to.be.an('Object');

					expect(msg.data['test-req-form:readval'].read).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].read.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].read.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].read.allowedTypes[0]).to.equal('string');
					expect(msg.data['test-req-form:readval'].read.isRequired).to.be.true;

					expect(msg.data['test-req-form:readval'].create).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].create.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].create.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].create.allowedTypes[0]).to.equal('object');
					expect(msg.data['test-req-form:readval'].create.isRequired).to.be.true;
					expect(msg.data['test-req-form:readval'].create.subValue).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].create.subValue['some-entry']).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].create.subValue['some-entry']).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].create.subValue['some-entry'].allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].create.subValue['some-entry'].allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].create.subValue['some-entry'].allowedTypes[0]).to.equal('email');
					expect(msg.data['test-req-form:readval'].create.subValue['some-entry'].isRequired).to.be.true;

					expect(msg.data['test-req-form:readval'].pw).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].pw.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].pw.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].pw.allowedTypes[0]).to.equal('any');
					expect(msg.data['test-req-form:readval'].pw.isRequired).to.be.false;

					expect(msg.data['test-req-form:readval'].complex).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].complex.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].complex.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].complex.allowedTypes[0]).to.equal('object');
					expect(msg.data['test-req-form:readval'].complex.isRequired).to.be.true;
					expect(msg.data['test-req-form:readval'].complex.subValue).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].complex.subValue.a).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].complex.subValue.a.allowedTypes[0]).to.equal('object');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.isRequired).to.be.true;
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b.allowedTypes[0]).to.equal('object');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b.isRequired).to.be.true;
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b.subValue).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b.subValue.c).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b.subValue.c.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b.subValue.c.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b.subValue.c.allowedTypes[0]).to.equal('string');
					expect(msg.data['test-req-form:readval'].complex.subValue.a.subValue.b.subValue.c.isRequired).to.be.true;
					expect(msg.data['test-req-form:readval'].complex.subValue.d).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].complex.subValue.d.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].complex.subValue.d.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].complex.subValue.d.allowedTypes[0]).to.equal('object');
					expect(msg.data['test-req-form:readval'].complex.subValue.d.isRequired).to.be.true;
					expect(msg.data['test-req-form:readval'].complex.subValue.d.subValue).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].complex.subValue.d.subValue.e).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].complex.subValue.d.subValue.e.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].complex.subValue.d.subValue.e.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].complex.subValue.d.subValue.e.allowedTypes[0]).to.equal('string');
					expect(msg.data['test-req-form:readval'].complex.subValue.d.subValue.e.isRequired).to.be.true;

					expect(msg.data['test-req-form:readval'].validator).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].validator.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].validator.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].validator.allowedTypes[0]).to.equal('object');
					expect(msg.data['test-req-form:readval'].validator.isRequired).to.be.true;
					expect(msg.data['test-req-form:readval'].validator.subValue).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].validator.subValue.a).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].validator.subValue.a.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].validator.subValue.a.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].validator.subValue.a.allowedTypes[0]).to.equal('object');
					expect(msg.data['test-req-form:readval'].validator.subValue.a.isRequired).to.be.true;
					expect(msg.data['test-req-form:readval'].validator.subValue.a.subValue).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].validator.subValue.a.subValue.b).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].validator.subValue.a.subValue.b.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].validator.subValue.a.subValue.b.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].validator.subValue.a.subValue.b.allowedTypes[0]).to.equal('any');
					expect(msg.data['test-req-form:readval'].validator.subValue.a.subValue.b.isRequired).to.be.false;
					expect(msg.data['test-req-form:readval'].validator.subValue.c).to.be.an('Object');
					expect(msg.data['test-req-form:readval'].validator.subValue.c.allowedTypes).to.be.an('Array');
					expect(msg.data['test-req-form:readval'].validator.subValue.c.allowedTypes.length).to.equal(1);
					expect(msg.data['test-req-form:readval'].validator.subValue.c.allowedTypes[0]).to.equal('boolean');
					expect(msg.data['test-req-form:readval'].validator.subValue.c.isRequired).to.be.true;

					resolveTestPromise();
				} catch(e) {
					rejectTestPromise(e);
				}
            }
        )

        await testPromise
	})
	it( 'server can be notified about updated api' )


    it( 'plugins are called with provided staticEndpointArguments', async function ()
    {
        const	staticArg       = 'Hallo'
        const   server          = new Yoctopus({ port, logInstructions:{}, staticEndpointArguments: { 'testStatArgs': staticArg } })

        let     openResult     = null
        let     endpointResult = null


        await server.plugin({
            name        : 'test-plugin'
        ,   listeners   : {
                            server: {
                                onConnectionOpened      : function({ testStatArgs })
                                                        {
                                                            openResult = testStatArgs
                                                        }
                            }
                        }
		,   endpoints   : {
                            echo: {
                                update: {
                                    handler         : ({ testStatArgs }) =>
                                                    {
                                                        endpointResult = testStatArgs
                                                    }
                                }
                            }
			}
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   async function open()
            {
                await pause( 0.1 )

				expect( openResult ).to.equal( staticArg )

				wsA.send(
                    JSON.stringify({
                        meta: {
                            id          : 'testMsgId-' + Math.random()
                        ,   method      : 'update'
                        ,   endpoint    : 'test-plugin:echo'
                        }
                    })
                )

				await pause( 0.1 )

				expect( endpointResult ).to.equal( staticArg )

                await server.close()

                resolveTestPromise()
            }
        )

        await testPromise
    })


    it( 'passes request data to endpoint handlers', async function ()
    {
        const	testValue           = Math.random()
        let     endpointReceived    = null

        await server.plugin({
            name        : 'test-plugin'
		,   endpoints: {
				echo: {
					update: {
                        handler : ( { request } ) =>
                                {
                                    const { meta, data } = request
                                    endpointReceived = data.testValue
                                }
					}
				}
			}
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   async function open()
            {
				wsA.send(
                    JSON.stringify({
                        meta: {
                                id          : 'testMsgId-' + Math.random()
                            ,   method      : 'update'
                            ,   endpoint    : 'test-plugin:echo'
                        }
                    ,   data: { testValue }
                    })
                )

				await pause( 0.1 )

				expect( endpointReceived ).to.equal( testValue )

                resolveTestPromise()
            }
        )

        await testPromise
    })




    it ( 'can send notifications', async function()
    {
       const testValue = Math.random()

        await server.plugin({
            name        : 'test-plugin1'
		,	listeners   : {
                            server: {
                                onConnectionOpened: function( { notify } )
                                {
                                    notify({ type: 'testnotification', data: testValue })
                                }
                            }
						}
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'message'
        ,   async function receive( textMsg )
            {
                const msg = JSON.parse( textMsg )

				expect( msg.data ).to.equal( testValue )

                resolveTestPromise()
            }
        )

        await testPromise
    })



    it ( 'can listen to connection setting changes', async function()
    {
        const	testValue   = Math.random()
        let     callCount   = 0

        const   checkCalls  = function({ value, setting, connectionId })
                            {
                                // console.log( 'called', setting, connectionId )
                                if ( setting === 'test-plugin-B.some-setting' )
                                {
                                    expect( value ).to.equal( testValue )
                                    callCount++
                                    if ( callCount === 3)
                                    {
                                        resolveTestPromise()
                                    }
                                }
                            }

        await server.plugin({
            name        : 'test-plugin-A'
		,   listeners	: {
                            plugins: {
                                'test-plugin-B': {
                                    'some-setting': checkCalls
                                ,   '*': checkCalls
                                ,   'some-other-setting': function( value )
                                    {
                                        throw new Error( 'I should not have been called!' )
                                    }
                                }
                            ,   '*' : checkCalls
                            }
                        }
        })

        await server.plugin({
            name        : 'test-plugin-B'
		,	settings    : {
                            'some-setting': 'number'
                        }
		,	endpoints	: {
                            'thevalue': {
                                update: {
                                    handler: ( { request, connectionSettings } ) =>
                                    {
                                        const { meta, data } = request
                                        connectionSettings.set( 'some-setting', request.data )
                                    }
                                }
                            }
						}
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   async function open()
            {
				wsA.send(
                    JSON.stringify({
                        meta: {
                                id          : 'testMsgId-' + Math.random()
                            ,   method      : 'update'
                            ,   endpoint    : 'test-plugin-B:thevalue'
                        }
                    ,   data: testValue
                    })
                )
            }
        )

        await testPromise
    })



    it ( 'can export a connection\'s store', async () =>
    {
        const	testValue   = Math.random()

        await server.plugin({
            name        : 'test-plugin'
        ,   settings    : {
                            'sett': 'any'
                        }
        ,   endpoints   : {
                            'setsett': {
                                update: {
                                    handler: ( { request, connectionSettings } ) =>
                                    {
                                        connectionSettings.set( 'sett', request.data )
                                    }
                                }
                            }
                        }
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   async function open()
            {
                wsA.send(
                    JSON.stringify({
                        meta: {
                                id          : 'testMsgId-' + Math.random()
                            ,   method      : 'update'
                            ,   endpoint    : 'test-plugin:setsett'
                        }
                    ,   data: testValue
                    })
                )

                await pause( 0.1 )

                const store = Object.values( server.connectionRegister )[ 0 ].store

                // console.log( store )

                expect( store ).to.be.an( 'object' )
                expect( store[ 'test-plugin' ] ).to.be.an( 'object' )
                expect( store[ 'test-plugin' ].sett ).to.be.a( 'number' )
                expect( store[ 'test-plugin' ].sett ).to.equal( testValue )

                resolveTestPromise()
            }
        )

        await testPromise
    })



    it ( 'has a logger for every plugin', async () =>
    {
        await server.close()

        const   logsFromInstruction    = []
        const   logsFromListener       = []

        server                          = new Yoctopus({ port, logInstructions : { info: ( log => logsFromInstruction.push( log )) } })

        const	testValue               = Math.random()

        server.subscribeToLogs( function( logMsg )
        {
            logsFromListener.push( logMsg )
        })

        await server.plugin({
            name        : 'test-plugin'
        ,   listeners   : {
                            'server' : {
                                onConnectionOpened: ( { log } ) =>
                                {
                                    // console.log( 'open', log )
                                    log.notice( 'opened' )
                                }
                            }
                        }
        ,   endpoints   : {
                            'setsett': {
                                update: {
                                    handler: ( { log } ) =>
                                    {
                                        // console.log( log )
                                        log.notice( 'ep called' )
                                    }
                                }
                            }
                        }
        })


        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   async function open()
            {
				try {
					wsA.send(
						JSON.stringify({
							meta: {
									id          : 'testMsgId-' + Math.random()
								,   method      : 'update'
								,   endpoint    : 'test-plugin:setsett'
							}
						,   data: testValue
						})
					)

					await pause( 0.1 )

					expect( logsFromInstruction.length ).to.equal( logsFromListener.length + 1 ) // internal plugin install log

					const   pluginLogs      = logsFromInstruction.filter( l => l.source === 'test-plugin' )

					expect( pluginLogs[ 0 ].source ).to.equal( 'test-plugin' )
					expect( pluginLogs[ 0 ].data ).to.equal( 'opened' )
					expect( pluginLogs[ 1 ].source ).to.equal( 'test-plugin' )
					expect( pluginLogs[ 1 ].data ).to.equal( 'ep called' )

					resolveTestPromise()
				} catch(e) {
					rejectTestPromise(e);
				}
            }
        )

        await testPromise
    })


    it ( 'can let a plugin add to the listener arguments', async () =>
    {
        const	testValue   = Math.random()

        await server.plugin({
            name                : 'test-plugin'
        ,   passToListeners     : {
                                    testValue
                                }
        ,   endpoints   :       {
                                    'someep': {
                                        read: {
                                            handler: ( { request, connectionSettings, 'test-plugin' : tp } ) =>
                                            {
                                                expect( tp.testValue ).to.equal( testValue )
                                                resolveTestPromise()
                                            }
                                        }
                                    }
                                }
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   async function open()
            {
                wsA.send(
                    JSON.stringify({
                        meta: {
                                id          : 'testMsgId-' + Math.random()
                            ,   method      : 'read'
                            ,   endpoint    : 'test-plugin:someep'
                        }
                    })
                )
            }
        )

        await testPromise
    })



    it ( 'can let a plugin add to the listener arguments', async () =>
    {
        const	testValue   = Math.random()

        await server.plugin({
            name                : 'test-plugin'
        ,   passToListeners     : {
                                    testValue
                                }
        ,   endpoints   :       {
                                    'someep': {
                                        read: {
                                            handler: ( { request, connectionSettings, 'test-plugin' : tp } ) =>
                                            {
                                                expect( tp.testValue ).to.equal( testValue )
                                                resolveTestPromise()
                                            }
                                        }
                                    }
                                }
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   async function open()
            {
                wsA.send(
                    JSON.stringify({
                        meta: {
                                id          : 'testMsgId-' + Math.random()
                            ,   method      : 'read'
                            ,   endpoint    : 'test-plugin:someep'
                        }
                    })
                )
            }
        )

        await testPromise
    })

    it ( 'can multicast', async () =>
    {
        let count = 0

        let msgReceivedCount = 0

        await server.plugin( server.bundledPlugins.multicast( server ) )

        await server.plugin({
            name                : 'notrel'
        ,   settings: {
                cast: 'bool'
            }
        ,   listeners : {
                server: {
                    onConnectionOpened: function({ connectionSettings })
                    {
                        count += 1

                        if ( count < 5 )
                        {
                            connectionSettings.set( 'cast', true )
                        }
                        else
                        {
                            connectionSettings.set( 'cast', false )
                        }
                    }
                }
            }
        ,   endpoints   :       {
                                    'docast': {
                                        read: {
                                            handler: ({ request, multicast }) =>
                                            {
                                                multicast({ notrel: { cast: true }}, { hi: 'yo' })
                                            }
                                        }
                                    }
                                }
        })

        await server.start()

        const connections = []

        for ( let i = 0 ; i < 10 ; i += 1 )
        {
            connections[ i ] = new WebSocket( `ws://localhost:${port}` )
            connections[ i ].on(
                'message'
            ,   function( msg )
                {
                    // console.log( msg )
                    const msgObj = JSON.parse( msg )
                    if ( msgObj.meta.id === 'test1' )
                    {
                        expect( msgReceivedCount ).to.equal( 4 )
                        resolveTestPromise()
                    }
                    else
                    {
                        msgReceivedCount += 1
                    }
                }
            )
            // connections[ i ].on( 'message', console.log )
        }

        connections[ 9 ]
            .on(
                'open'
            ,   function()
                {
                    connections[ 9 ].send(
                        JSON.stringify({
                            meta: { id: 'test1', method: 'read', endpoint: 'notrel:docast' }
                        })
                    )
                }
            )

        await testPromise
    })


    it ( 'has an internal plugin', async () =>
    {
        const	testValue   = Math.random()

        await server.plugin({
            name                : 'test-plugin'
        ,   listeners           : {
                                    plugins: {
                                        yoctopus: {
                                            label: function({ value, setting })
                                            {
                                                if ( setting === 'yoctopus.label' )
                                                {
                                                    expect( value ).to.equal( String( testValue ) )
                                                    resolveTestPromise()
                                                }
                                            }
                                        }
                                    }
                                }
        })

        await server.start()

        wsA = new WebSocket( `ws://localhost:${port}` )

        wsA.on(
            'open'
        ,   async function open()
            {
                wsA.send(
                    JSON.stringify({
                        meta: {
                                id          : 'testMsgId-' + Math.random()
                            ,   method      : 'update'
                            ,   endpoint    : 'yoctopus:label'
                        }
                    ,   data: String( testValue )
                    })
                )
            }
        )

        await testPromise
    })


    it ( 'closes connections that don\'t respond to ping', async () =>
    {
        const   pingFailAmount  = 3
        const   pingTime        = 250
        const   silentAfter     = 5
        let     numPings        = 0

		server = new Yoctopus({ port, pingTime, pingFailAmount })

        console.log( 'Starting server' )

		await server.start()

		wsA = new WebSocket( `ws://localhost:${port}` )
        
        wsA.on( 'open', function()
        {
            console.log( 'ws opened' )

            wsA._receiver.removeAllListeners( 'ping' )

            wsA._receiver.on( 'ping', function()
            {
                numPings++
                console.log( `Websocket received ping ${numPings}` )

                if ( numPings <= silentAfter )
                {
                    console.log( 'Explicitly sending pong' )
                    wsA.pong()

                    if ( numPings === silentAfter )
                    {
                        //disable sending pong on ping therby breaking the spec but we simulate a dead connection this way
                        //instead just count the number of pings received

                        console.log( 'Removing ping listener' )
                    }
                }
                else
                {
                    console.log( 'Ignoring ping' )
                }
            })
            
            wsA.on( 'close', function()
            {
                console.log( 'Websocket closed by other side' )

                try
                {
                    expect( numPings ).to.equal( pingFailAmount + silentAfter )
					resolveTestPromise()
                }
                catch( e )
                {
					rejectTestPromise( e )
				}
			})
		})

		await testPromise
	})
})

describe( 'origin check', function()
{
    let server              = null
    let port                = null
    let ws                  = null

    beforeEach( async function()
    {
        port    = nextPort()
        server  = new Yoctopus({ port, logInstructions:{} }) // no * default, so it can be set
    })

    afterEach( async function()
    {
        try { await server.close() } catch( err ) {}
        try { await ws.close() } catch( err ) {}
    })

    for ( let ori of [ [ 'ws://google' ], '' ] )
    {
        it ( `closes connection if only origin [${ori}] is allowed`, async () =>
        {
            server.allowedOrigins = [ '' ]

            await server.start()

            ws = new WebSocket( `ws://localhost:${port}` )

            await new Promise( function( resolve )
            {
                ws.on( 'close', () => {
                    console.log('close')
                    resolve()
                })
            })
        })
    }

    for ( let ori of [ [ 'ws://localhost' ], '*' ] )
    {
        it ( `allows connection if origin ${ori} is allowed`, async () =>
        {
            await server.start()
            server.allowedOrigins = ori

            ws = new WebSocket( `ws://localhost:${port}`, { origin: 'ws://localhost' } )

            let closed = false;
            let opened = false;
            await new Promise( function( resolve, reject )
            {
                setTimeout( resolve, 1000 );

                ws.on( 'close', () => {
                    console.log('close')
                    closed = true;
                })
                ws.on( 'open', () => {
                    console.log('open')
                    opened = true;
                })
            })
        })
    }
})
